Class: Net::HTTP::Persistent

Inherits:
Object
  • Object
show all
Defined in:
lib/net/http/persistent.rb

Overview

Persistent connections for Net::HTTP

Net::HTTP::Persistent maintains persistent connections across all the servers you wish to talk to. For each host:port you communicate with a single persistent connection is created.

Multiple Net::HTTP::Persistent objects will share the same set of connections.

For each thread you start a new connection will be created. A Net::HTTP::Persistent connection will not be shared across threads.

You can shut down the HTTP connections when done by calling #shutdown. You should name your Net::HTTP::Persistent object if you intend to call this method.

Example:

require 'net/http/persistent'

uri = URI 'http://example.com/awesome/web/service'

http = Net::HTTP::Persistent.new 'my_app_name'

# perform a GET
response = http.request uri

# or

get = Net::HTTP::Get.new uri.request_uri
response = http.request get

# create a POST
post_uri = uri + 'create'
post = Net::HTTP::Post.new post_uri.path
post.set_form_data 'some' => 'cool data'

# perform the POST, the URI is always required
response http.request post_uri, post

Note that for GET, HEAD and other requests that do not have a body you want to use URI#request_uri not URI#path. The request_uri contains the query params which are sent in the body for other requests.

SSL

SSL connections are automatically created depending upon the scheme of the URI. SSL connections are automatically verified against the default certificate store for your computer. You can override this by changing verify_mode or by specifying an alternate cert_store.

Here are the SSL settings, see the individual methods for documentation:

#certificate

This client’s certificate

#ca_file

The certificate-authorities

#ca_path

Directory with certificate-authorities

#cert_store

An SSL certificate store

#ciphers

List of SSl ciphers allowed

#private_key

The client’s SSL private key

#reuse_ssl_sessions

Reuse a previously opened SSL session for a new connection

#ssl_timeout

SSL session lifetime

#ssl_version

Which specific SSL version to use

#verify_callback

For server certificate verification

#verify_depth

Depth of certificate verification

#verify_mode

How connections should be verified

Proxies

A proxy can be set through #proxy= or at initialization time by providing a second argument to ::new. The proxy may be the URI of the proxy server or :ENV which will consult environment variables.

See #proxy= and #proxy_from_env for details.

Headers

Headers may be specified for use in every request. #headers are appended to any headers on the request. #override_headers replace existing headers on the request.

The difference between the two can be seen in setting the User-Agent. Using http.headers['User-Agent'] = 'MyUserAgent' will send “Ruby, MyUserAgent” while http.override_headers['User-Agent'] = 'MyUserAgent' will send “MyUserAgent”.

Tuning

Segregation

By providing an application name to ::new you can separate your connections from the connections of other applications.

Idle Timeout

If a connection hasn’t been used for this number of seconds it will automatically be reset upon the next use to avoid attempting to send to a closed connection. The default value is 5 seconds. nil means no timeout. Set through #idle_timeout.

Reducing this value may help avoid the “too many connection resets” error when sending non-idempotent requests while increasing this value will cause fewer round-trips.

Read Timeout

The amount of time allowed between reading two chunks from the socket. Set through #read_timeout

Max Requests

The number of requests that should be made before opening a new connection. Typically many keep-alive capable servers tune this to 100 or less, so the 101st request will fail with ECONNRESET. If unset (default), this value has no effect, if set, connections will be reset on the request after max_requests.

Open Timeout

The amount of time to wait for a connection to be opened. Set through #open_timeout.

Socket Options

Socket options may be set on newly-created connections. See #socket_options for details.

Non-Idempotent Requests

By default non-idempotent requests will not be retried per RFC 2616. By setting retry_change_requests to true requests will automatically be retried once.

Only do this when you know that retrying a POST or other non-idempotent request is safe for your application and will not create duplicate resources.

The recommended way to handle non-idempotent requests is the following:

require 'net/http/persistent'

uri = URI 'http://example.com/awesome/web/service'
post_uri = uri + 'create'

http = Net::HTTP::Persistent.new 'my_app_name'

post = Net::HTTP::Post.new post_uri.path
# ... fill in POST request

begin
  response = http.request post_uri, post
rescue Net::HTTP::Persistent::Error

  # POST failed, make a new request to verify the server did not process
  # the request
  exists_uri = uri + '...'
  response = http.get exists_uri

  # Retry if it failed
  retry if response.code == '404'
end

The method of determining if the resource was created or not is unique to the particular service you are using. Of course, you will want to add protection from infinite looping.

Connection Termination

If you are done using the Net::HTTP::Persistent instance you may shut down all the connections in the current thread with #shutdown. This is not recommended for normal use, it should only be used when it will be several minutes before you make another HTTP request.

If you are using multiple threads, call #shutdown in each thread when the thread is done making requests. If you don’t call shutdown, that’s OK. Ruby will automatically garbage collect and shutdown your HTTP connections when the thread terminates.

Defined Under Namespace

Classes: Error

Constant Summary collapse

EPOCH =

The beginning of Time

Time.at 0
HAVE_OPENSSL =

Is OpenSSL available? This test works with autoload

defined? OpenSSL::SSL # :nodoc:
VERSION =

The version of Net::HTTP::Persistent you are using

'2.10.0'
RETRIED_EXCEPTIONS =

Exceptions rescued for automatic retry on ruby 2.0.0. This overlaps with the exception list for ruby 1.x.

[ # :nodoc:
  (Net::ReadTimeout if Net.const_defined? :ReadTimeout),
  IOError,
  EOFError,
  Errno::ECONNRESET,
  Errno::ECONNABORTED,
  Errno::EPIPE,
  (OpenSSL::SSL::SSLError if HAVE_OPENSSL),
  Timeout::Error,
].compact

Instance Attribute Summary collapse

Class Method Summary collapse

Instance Method Summary collapse

Constructor Details

#initialize(name = nil, proxy = nil) ⇒ Persistent

Creates a new Net::HTTP::Persistent.

Set name to keep your connections apart from everybody else’s. Not required currently, but highly recommended. Your library name should be good enough. This parameter will be required in a future version.

proxy may be set to a URI::HTTP or :ENV to pick up proxy options from the environment. See proxy_from_env for details.

In order to use a URI for the proxy you may need to do some extra work beyond URI parsing if the proxy requires a password:

proxy = URI 'http://proxy.example'
proxy.user     = 'AzureDiamond'
proxy.password = 'hunter2'


508
509
510
511
512
513
514
515
516
517
518
519
520
521
522
523
524
525
526
527
528
529
530
531
532
533
534
535
536
537
538
539
540
541
542
543
544
545
546
547
548
549
550
551
552
553
554
555
556
557
558
559
560
561
562
563
# File 'lib/net/http/persistent.rb', line 508

def initialize name = nil, proxy = nil
  @name = name

  @debug_output     = nil
  @proxy_uri        = nil
  @no_proxy         = []
  @headers          = {}
  @override_headers = {}
  @http_versions    = {}
  @keep_alive       = 30
  @open_timeout     = nil
  @read_timeout     = nil
  @idle_timeout     = 5
  @max_requests     = nil
  @socket_options   = []
  @ssl_generation   = 0 # incremented when SSL session variables change

  @socket_options << [Socket::IPPROTO_TCP, Socket::TCP_NODELAY, 1] if
    Socket.const_defined? :TCP_NODELAY

  key = ['net_http_persistent', name].compact
  @generation_key     = [key, 'generations'    ].join('_').intern
  @ssl_generation_key = [key, 'ssl_generations'].join('_').intern
  @timeout_key        = [key, 'timeouts'       ].join('_').intern

  pool_size = Process.getrlimit(Process::RLIMIT_NOFILE).first / 4
  @pool     = Net::HTTP::Persistent::Pool.new size: pool_size do |http_args|
    Net::HTTP::Persistent::Connection.new http_class, http_args, @ssl_generation
  end

  @certificate        = nil
  @ca_file            = nil
  @ca_path            = nil
  @ciphers            = nil
  @private_key        = nil
  @ssl_timeout        = nil
  @ssl_version        = nil
  @verify_callback    = nil
  @verify_depth       = nil
  @verify_mode        = nil
  @cert_store         = nil

  @generation         = 0 # incremented when proxy URI changes

  if HAVE_OPENSSL then
    @verify_mode        = OpenSSL::SSL::VERIFY_PEER
    @reuse_ssl_sessions = OpenSSL::SSL.const_defined? :Session
  end

  @retry_change_requests = false

  @ruby_1 = RUBY_VERSION < '2'
  @retried_on_ruby_2 = !@ruby_1

  self.proxy = proxy if proxy
end

Instance Attribute Details

#ca_fileObject

An SSL certificate authority. Setting this will set verify_mode to VERIFY_PEER.



296
297
298
# File 'lib/net/http/persistent.rb', line 296

def ca_file
  @ca_file
end

#ca_pathObject

A directory of SSL certificates to be used as certificate authorities. Setting this will set verify_mode to VERIFY_PEER.



302
303
304
# File 'lib/net/http/persistent.rb', line 302

def ca_path
  @ca_path
end

#cert_storeObject

An SSL certificate store. Setting this will override the default certificate store. See verify_mode for more information.



308
309
310
# File 'lib/net/http/persistent.rb', line 308

def cert_store
  @cert_store
end

#certificateObject Also known as: cert

This client’s OpenSSL::X509::Certificate



287
288
289
# File 'lib/net/http/persistent.rb', line 287

def certificate
  @certificate
end

#ciphersObject

The ciphers allowed for SSL connections



313
314
315
# File 'lib/net/http/persistent.rb', line 313

def ciphers
  @ciphers
end

#debug_outputObject

Sends debug_output to this IO via Net::HTTP#set_debug_output.

Never use this method in production code, it causes a serious security hole.



321
322
323
# File 'lib/net/http/persistent.rb', line 321

def debug_output
  @debug_output
end

#generationObject (readonly)

Current connection generation



326
327
328
# File 'lib/net/http/persistent.rb', line 326

def generation
  @generation
end

#generation_keyObject (readonly)

Where this instance’s connections live in the thread local variables



331
332
333
# File 'lib/net/http/persistent.rb', line 331

def generation_key
  @generation_key
end

#headersObject (readonly)

Headers that are added to every request using Net::HTTP#add_field



336
337
338
# File 'lib/net/http/persistent.rb', line 336

def headers
  @headers
end

#http_versionsObject (readonly)

Maps host:port to an HTTP version. This allows us to enable version specific features.



342
343
344
# File 'lib/net/http/persistent.rb', line 342

def http_versions
  @http_versions
end

#idle_timeoutObject

Maximum time an unused connection can remain idle before being automatically closed.



348
349
350
# File 'lib/net/http/persistent.rb', line 348

def idle_timeout
  @idle_timeout
end

#keep_aliveObject

The value sent in the Keep-Alive header. Defaults to 30. Not needed for HTTP/1.1 servers.

This may not work correctly for HTTP/1.0 servers

This method may be removed in a future version as RFC 2616 does not require this header.



365
366
367
# File 'lib/net/http/persistent.rb', line 365

def keep_alive
  @keep_alive
end

#max_requestsObject

Maximum number of requests on a connection before it is considered expired and automatically closed.



354
355
356
# File 'lib/net/http/persistent.rb', line 354

def max_requests
  @max_requests
end

#nameObject (readonly)

A name for this connection. Allows you to keep your connections apart from everybody else’s.



371
372
373
# File 'lib/net/http/persistent.rb', line 371

def name
  @name
end

#no_proxyObject (readonly)

List of host suffixes which will not be proxied



399
400
401
# File 'lib/net/http/persistent.rb', line 399

def no_proxy
  @no_proxy
end

#open_timeoutObject

Seconds to wait until a connection is opened. See Net::HTTP#open_timeout



376
377
378
# File 'lib/net/http/persistent.rb', line 376

def open_timeout
  @open_timeout
end

#override_headersObject (readonly)

Headers that are added to every request using Net::HTTP#[]=



381
382
383
# File 'lib/net/http/persistent.rb', line 381

def override_headers
  @override_headers
end

#poolObject (readonly)

Test-only accessor for the connection pool



404
405
406
# File 'lib/net/http/persistent.rb', line 404

def pool
  @pool
end

#private_keyObject Also known as: key

This client’s SSL private key



386
387
388
# File 'lib/net/http/persistent.rb', line 386

def private_key
  @private_key
end

#proxy_uriObject (readonly)

The URL through which requests will be proxied



394
395
396
# File 'lib/net/http/persistent.rb', line 394

def proxy_uri
  @proxy_uri
end

#read_timeoutObject

Seconds to wait until reading one block. See Net::HTTP#read_timeout



409
410
411
# File 'lib/net/http/persistent.rb', line 409

def read_timeout
  @read_timeout
end

#retry_change_requestsObject

Enable retries of non-idempotent requests that change data (e.g. POST requests) when the server has disconnected.

This will in the worst case lead to multiple requests with the same data, but it may be useful for some applications. Take care when enabling this option to ensure it is safe to POST or perform other non-idempotent requests to the server.



489
490
491
# File 'lib/net/http/persistent.rb', line 489

def retry_change_requests
  @retry_change_requests
end

#reuse_ssl_sessionsObject

By default SSL sessions are reused to avoid extra SSL handshakes. Set this to false if you have problems communicating with an HTTPS server like:

SSL_connect [...] read finished A: unexpected message (OpenSSL::SSL::SSLError)


418
419
420
# File 'lib/net/http/persistent.rb', line 418

def reuse_ssl_sessions
  @reuse_ssl_sessions
end

#socket_optionsObject (readonly)

An array of options for Socket#setsockopt.

By default the TCP_NODELAY option is set on sockets.

To set additional options append them to this array:

http.socket_options << [Socket::SOL_SOCKET, Socket::SO_KEEPALIVE, 1]


429
430
431
# File 'lib/net/http/persistent.rb', line 429

def socket_options
  @socket_options
end

#ssl_generationObject (readonly)

Current SSL connection generation



434
435
436
# File 'lib/net/http/persistent.rb', line 434

def ssl_generation
  @ssl_generation
end

#ssl_generation_keyObject (readonly)

Where this instance’s SSL connections live in the thread local variables



439
440
441
# File 'lib/net/http/persistent.rb', line 439

def ssl_generation_key
  @ssl_generation_key
end

#ssl_timeoutObject

SSL session lifetime



444
445
446
# File 'lib/net/http/persistent.rb', line 444

def ssl_timeout
  @ssl_timeout
end

#ssl_versionObject

SSL version to use.

By default, the version will be negotiated automatically between client and server. Ruby 1.9 and newer only.



452
453
454
# File 'lib/net/http/persistent.rb', line 452

def ssl_version
  @ssl_version
end

#timeout_keyObject (readonly)

Where this instance’s last-use times live in the thread local variables



457
458
459
# File 'lib/net/http/persistent.rb', line 457

def timeout_key
  @timeout_key
end

#verify_callbackObject

SSL verification callback. Used when ca_file or ca_path is set.



462
463
464
# File 'lib/net/http/persistent.rb', line 462

def verify_callback
  @verify_callback
end

#verify_depthObject

Sets the depth of SSL certificate verification



467
468
469
# File 'lib/net/http/persistent.rb', line 467

def verify_depth
  @verify_depth
end

#verify_modeObject

HTTPS verify mode. Defaults to OpenSSL::SSL::VERIFY_PEER which verifies the server certificate.

If no ca_file, ca_path or cert_store is set the default system certificate store is used.

You can use verify_mode to override any default values.



478
479
480
# File 'lib/net/http/persistent.rb', line 478

def verify_mode
  @verify_mode
end

Class Method Details

.detect_idle_timeout(uri, max = 10) ⇒ Object

Use this method to detect the idle timeout of the host at uri. The value returned can be used to configure #idle_timeout. max controls the maximum idle timeout to detect.

After

Idle timeout detection is performed by creating a connection then performing a HEAD request in a loop until the connection terminates waiting one additional second per loop.

NOTE: This may not work on ruby > 1.9.



248
249
250
251
252
253
254
255
256
257
258
259
260
261
262
263
264
265
266
267
268
269
270
271
272
273
274
275
276
277
278
279
280
281
282
# File 'lib/net/http/persistent.rb', line 248

def self.detect_idle_timeout uri, max = 10
  uri = URI uri unless URI::Generic === uri
  uri += '/'

  req = Net::HTTP::Head.new uri.request_uri

  http = new 'net-http-persistent detect_idle_timeout'

  http.connection_for uri do |connection|
    sleep_time = 0

    http = connection.http

    loop do
      response = http.request req

      $stderr.puts "HEAD #{uri} => #{response.code}" if $DEBUG

      unless Net::HTTPOK === response then
        raise Error, "bad response code #{response.code} detecting idle timeout"
      end

      break if sleep_time >= max

      sleep_time += 1

      $stderr.puts "sleeping #{sleep_time}" if $DEBUG
      sleep sleep_time
    end
  end
rescue
  # ignore StandardErrors, we've probably found the idle timeout.
ensure
  return sleep_time unless $!
end

Instance Method Details

#can_retry?(req, retried_on_ruby_2 = false) ⇒ Boolean

Is the request req idempotent or is retry_change_requests allowed.

If retried_on_ruby_2 is true, true will be returned if we are on ruby, retry_change_requests is allowed and the request is not idempotent.

Returns:

  • (Boolean)


783
784
785
786
787
# File 'lib/net/http/persistent.rb', line 783

def can_retry? req, retried_on_ruby_2 = false
  return @retry_change_requests && !idempotent?(req) if retried_on_ruby_2

  @retry_change_requests || idempotent?(req)
end

#cleanup(generation, thread = Thread.current, generation_key = @generation_key) ⇒ Object

Finishes all connections on the given thread that were created before the given generation in the threads generation_key list.

See #shutdown for a bunch of scary warning about misusing this method.



620
621
622
623
624
625
626
627
628
629
630
631
632
633
634
635
# File 'lib/net/http/persistent.rb', line 620

def cleanup(generation, thread = Thread.current,
            generation_key = @generation_key) # :nodoc:
  timeouts = thread[@timeout_key]

  (0...generation).each do |old_generation|
    next unless thread[generation_key]

    conns = thread[generation_key].delete old_generation

    conns.each_value do |conn|
      finish conn, thread

      timeouts.delete conn.object_id if timeouts
    end if conns
  end
end

#connection_close?(header) ⇒ Boolean

Workaround for missing Net::HTTPRequest#connection_close? on Ruby 1.8

Returns:

  • (Boolean)


793
794
795
# File 'lib/net/http/persistent.rb', line 793

def connection_close? header
  header.connection_close?
end

#connection_for(uri) ⇒ Object

Creates a new connection for uri



640
641
642
643
644
645
646
647
648
649
650
651
652
653
654
655
656
657
658
659
660
661
662
663
664
665
666
667
668
669
670
671
672
673
674
675
676
677
678
679
680
# File 'lib/net/http/persistent.rb', line 640

def connection_for uri
  Thread.current[@timeout_key]        ||= Hash.new EPOCH

  use_ssl = uri.scheme.downcase == 'https'

  net_http_args = [uri.host, uri.port]

  net_http_args.concat @proxy_args if
    @proxy_uri and not proxy_bypass? uri.host, uri.port

  connection = @pool.checkout net_http_args

  http = connection.http

  connection.ressl @ssl_generation if
    connection.ssl_generation != @ssl_generation

  if not http.started? then
    ssl   http if use_ssl
    start http
  elsif expired? connection then
    reset connection
  end

  http.read_timeout = @read_timeout if @read_timeout
  http.keep_alive_timeout = @idle_timeout if @idle_timeout && http.respond_to?(:keep_alive_timeout=)

  return yield connection
rescue Errno::ECONNREFUSED
  address = http.proxy_address || http.address
  port    = http.proxy_port    || http.port

  raise Error, "connection refused: #{address}:#{port}"
rescue Errno::EHOSTDOWN
  address = http.proxy_address || http.address
  port    = http.proxy_port    || http.port

  raise Error, "host down: #{address}:#{port}"
ensure
  @pool.checkin net_http_args
end

#connection_keep_alive?(header) ⇒ Boolean

Workaround for missing Net::HTTPRequest#connection_keep_alive? on Ruby 1.8

Returns:

  • (Boolean)


800
801
802
# File 'lib/net/http/persistent.rb', line 800

def connection_keep_alive? header
  header.connection_keep_alive?
end

#error_message(connection) ⇒ Object

Returns an error message containing the number of requests performed on this connection



686
687
688
689
690
691
692
693
# File 'lib/net/http/persistent.rb', line 686

def error_message connection
  connection.requests -= 1 # fixup

  age = Time.now - connection.last_use

  "after #{connection.requests} requests on #{connection.http.object_id}, " \
    "last used #{age} seconds ago"
end

#escape(str) ⇒ Object

URI::escape wrapper



698
699
700
# File 'lib/net/http/persistent.rb', line 698

def escape str
  CGI.escape str if str
end

#expired?(connection) ⇒ Boolean

Returns true if the connection should be reset due to an idle timeout, or maximum request count, false otherwise.

Returns:

  • (Boolean)


714
715
716
717
718
719
720
# File 'lib/net/http/persistent.rb', line 714

def expired? connection
  return true  if     @max_requests && connection.requests >= @max_requests
  return false unless @idle_timeout
  return true  if     @idle_timeout.zero?

  Time.now - connection.last_use > @idle_timeout
end

#finish(connection) ⇒ Object

Finishes the Net::HTTP connection



743
744
745
# File 'lib/net/http/persistent.rb', line 743

def finish connection
  connection.finish
end

#http_classObject

:nodoc:



747
748
749
750
751
752
753
754
755
756
757
# File 'lib/net/http/persistent.rb', line 747

def http_class # :nodoc:
  if RUBY_VERSION > '2.0' then
    Net::HTTP
  elsif [:Artifice, :FakeWeb, :WebMock].any? { |klass|
           Object.const_defined?(klass)
        } or not @reuse_ssl_sessions then
      Net::HTTP
  else
    Net::HTTP::Persistent::SSLReuse
  end
end

#http_version(uri) ⇒ Object

Returns the HTTP protocol version for uri



762
763
764
# File 'lib/net/http/persistent.rb', line 762

def http_version uri
  @http_versions["#{uri.host}:#{uri.port}"]
end

#idempotent?(req) ⇒ Boolean

Is req idempotent according to RFC 2616?

Returns:

  • (Boolean)


769
770
771
772
773
774
775
# File 'lib/net/http/persistent.rb', line 769

def idempotent? req
  case req
  when Net::HTTP::Delete, Net::HTTP::Get, Net::HTTP::Head,
       Net::HTTP::Options, Net::HTTP::Put, Net::HTTP::Trace then
    true
  end
end

#max_ageObject

Deprecated in favor of #expired?



824
825
826
827
828
# File 'lib/net/http/persistent.rb', line 824

def max_age # :nodoc:
  return Time.now + 1 unless @idle_timeout

  Time.now - @idle_timeout
end

#normalize_uri(uri) ⇒ Object

Adds “http://” to the String uri if it is missing.



833
834
835
# File 'lib/net/http/persistent.rb', line 833

def normalize_uri uri
  (uri =~ /^https?:/) ? uri : "http://#{uri}"
end

#pipeline(uri, requests, &block) ⇒ Object

Pipelines requests to the HTTP server at uri yielding responses if a block is given. Returns all responses recieved.

See Net::HTTP::Pipeline for further details.

Only if net-http-pipeline was required before net-http-persistent #pipeline will be present.



848
849
850
851
852
# File 'lib/net/http/persistent.rb', line 848

def pipeline uri, requests, &block # :yields: responses
  connection_for uri do |connection|
    connection.http.pipeline requests, &block
  end
end

#proxy=(proxy) ⇒ Object

Sets the proxy server. The proxy may be the URI of the proxy server, the symbol :ENV which will read the proxy from the environment or nil to disable use of a proxy. See #proxy_from_env for details on setting the proxy from the environment.

If the proxy URI is set after requests have been made, the next request will shut-down and re-open all connections.

The no_proxy query parameter can be used to specify hosts which shouldn’t be reached via proxy; if set it should be a comma separated list of hostname suffixes, optionally with :port appended, for example example.com,some.host:8080.



880
881
882
883
884
885
886
887
888
889
890
891
892
893
894
895
896
897
898
899
900
901
902
903
904
905
906
907
# File 'lib/net/http/persistent.rb', line 880

def proxy= proxy
  @proxy_uri = case proxy
               when :ENV      then proxy_from_env
               when URI::HTTP then proxy
               when nil       then # ignore
               else raise ArgumentError, 'proxy must be :ENV or a URI::HTTP'
               end

  @no_proxy.clear

  if @proxy_uri then
    @proxy_args = [
      @proxy_uri.host,
      @proxy_uri.port,
      unescape(@proxy_uri.user),
      unescape(@proxy_uri.password),
    ]

    @proxy_connection_id = [nil, *@proxy_args].join ':'

    if @proxy_uri.query then
      @no_proxy = CGI.parse(@proxy_uri.query)['no_proxy'].join(',').downcase.split(',').map { |x| x.strip }.reject { |x| x.empty? }
    end
  end

  reconnect
  reconnect_ssl
end

#proxy_bypass?(host, port) ⇒ Boolean

Returns true when proxy should by bypassed for host.

Returns:

  • (Boolean)


954
955
956
957
958
959
960
961
962
963
964
# File 'lib/net/http/persistent.rb', line 954

def proxy_bypass? host, port
  host = host.downcase
  host_port = [host, port].join ':'

  @no_proxy.each do |name|
    return true if host[-name.length, name.length] == name or
       host_port[-name.length, name.length] == name
  end

  false
end

#proxy_from_envObject

Creates a URI for an HTTP proxy server from ENV variables.

If HTTP_PROXY is set a proxy will be returned.

If HTTP_PROXY_USER or HTTP_PROXY_PASS are set the URI is given the indicated user and password unless HTTP_PROXY contains either of these in the URI.

The NO_PROXY ENV variable can be used to specify hosts which shouldn’t be reached via proxy; if set it should be a comma separated list of hostname suffixes, optionally with :port appended, for example example.com,some.host:8080. When set to * no proxy will be returned.

For Windows users, lowercase ENV variables are preferred over uppercase ENV variables.



927
928
929
930
931
932
933
934
935
936
937
938
939
940
941
942
943
944
945
946
947
948
949
# File 'lib/net/http/persistent.rb', line 927

def proxy_from_env
  env_proxy = ENV['http_proxy'] || ENV['HTTP_PROXY']

  return nil if env_proxy.nil? or env_proxy.empty?

  uri = URI normalize_uri env_proxy

  env_no_proxy = ENV['no_proxy'] || ENV['NO_PROXY']

  # '*' is special case for always bypass
  return nil if env_no_proxy == '*'

  if env_no_proxy then
    uri.query = "no_proxy=#{escape(env_no_proxy)}"
  end

  unless uri.user or uri.password then
    uri.user     = escape ENV['http_proxy_user'] || ENV['HTTP_PROXY_USER']
    uri.password = escape ENV['http_proxy_pass'] || ENV['HTTP_PROXY_PASS']
  end

  uri
end

#reconnectObject

Forces reconnection of HTTP connections.



969
970
971
# File 'lib/net/http/persistent.rb', line 969

def reconnect
  @generation += 1
end

#reconnect_sslObject

Forces reconnection of SSL connections.



976
977
978
# File 'lib/net/http/persistent.rb', line 976

def reconnect_ssl
  @ssl_generation += 1
end

#request(uri, req = nil, &block) ⇒ Object

Makes a request on uri. If req is nil a Net::HTTP::Get is performed against uri.

If a block is passed #request behaves like Net::HTTP#request (the body of the response will not have been read).

req must be a Net::HTTPRequest subclass (see Net::HTTP for a list).

If there is an error and the request is idempotent according to RFC 2616 it will be retried automatically.



1011
1012
1013
1014
1015
1016
1017
1018
1019
1020
1021
1022
1023
1024
1025
1026
1027
1028
1029
1030
1031
1032
1033
1034
1035
1036
1037
1038
1039
1040
1041
1042
1043
1044
1045
1046
1047
1048
1049
1050
1051
1052
1053
1054
1055
1056
1057
1058
1059
1060
1061
1062
1063
1064
1065
1066
1067
1068
1069
# File 'lib/net/http/persistent.rb', line 1011

def request uri, req = nil, &block
  retried      = false
  bad_response = false

  req      = request_setup req || uri
  response = nil

  connection_for uri do |connection|
    http = connection.http

    begin
      connection.requests += 1

      response = http.request req, &block

      if connection_close?(req) or
         (response.http_version <= '1.0' and
          not connection_keep_alive?(response)) or
         connection_close?(response) then
        finish connection
      end
    rescue Net::HTTPBadResponse => e
      message = error_message connection

      finish connection

      raise Error, "too many bad responses #{message}" if
      bad_response or not can_retry? req

      bad_response = true
      retry
    rescue *RETRIED_EXCEPTIONS => e # retried on ruby 2
      request_failed e, req, connection if
        retried or not can_retry? req, @retried_on_ruby_2

      reset connection

      retried = true
      retry
    rescue Errno::EINVAL, Errno::ETIMEDOUT => e # not retried on ruby 2
      request_failed e, req, connection if retried or not can_retry? req

      reset connection

      retried = true
      retry
    rescue Exception => e
      finish connection

      raise
    ensure
      connection.last_use = Time.now
    end
  end

  @http_versions["#{uri.host}:#{uri.port}"] ||= response.http_version

  response
end

#request_failed(exception, req, connection) ⇒ Object

Raises an Error for exception which resulted from attempting the request req on the connection.

Finishes the connection.

Raises:



1077
1078
1079
1080
1081
1082
1083
1084
# File 'lib/net/http/persistent.rb', line 1077

def request_failed exception, req, connection # :nodoc:
  due_to = "(due to #{exception.message} - #{exception.class})"
  message = "too many connection resets #{due_to} #{error_message connection}"

  finish connection

  raise Error, message, exception.backtrace
end

#request_setup(req_or_uri) ⇒ Object

Creates a GET request if req_or_uri is a URI and adds headers to the request.

Returns the request.



1092
1093
1094
1095
1096
1097
1098
1099
1100
1101
1102
1103
1104
1105
1106
1107
1108
1109
1110
1111
1112
1113
# File 'lib/net/http/persistent.rb', line 1092

def request_setup req_or_uri # :nodoc:
  req = if URI === req_or_uri then
          Net::HTTP::Get.new req_or_uri.request_uri
        else
          req_or_uri
        end

  @headers.each do |pair|
    req.add_field(*pair)
  end

  @override_headers.each do |name, value|
    req[name] = value
  end

  unless req['Connection'] then
    req.add_field 'Connection', 'keep-alive'
    req.add_field 'Keep-Alive', @keep_alive
  end

  req
end

#reset(connection) ⇒ Object

Finishes then restarts the Net::HTTP connection



983
984
985
986
987
988
989
990
991
992
993
994
995
996
997
# File 'lib/net/http/persistent.rb', line 983

def reset connection
  http = connection.http

  finish connection

  start http
rescue Errno::ECONNREFUSED
  e = Error.new "connection refused: #{http.address}:#{http.port}"
  e.set_backtrace $@
  raise e
rescue Errno::EHOSTDOWN
  e = Error.new "host down: #{http.address}:#{http.port}"
  e.set_backtrace $@
  raise e
end

#shutdownObject

Shuts down all connections

NOTE: Calling shutdown for can be dangerous!

If any thread is still using a connection it may cause an error! Call #shutdown when you are completely done making requests!



1123
1124
1125
1126
1127
# File 'lib/net/http/persistent.rb', line 1123

def shutdown
  @pool.available.shutdown do |http|
    http.finish
  end
end

#ssl(connection) ⇒ Object

Enables SSL on connection



1132
1133
1134
1135
1136
1137
1138
1139
1140
1141
1142
1143
1144
1145
1146
1147
1148
1149
1150
1151
1152
1153
1154
1155
1156
1157
1158
1159
1160
1161
1162
1163
1164
1165
1166
1167
1168
1169
1170
1171
1172
1173
1174
1175
1176
1177
1178
1179
1180
1181
1182
1183
1184
1185
1186
1187
1188
1189
# File 'lib/net/http/persistent.rb', line 1132

def ssl connection
  connection.use_ssl = true

  connection.ciphers     = @ciphers     if @ciphers
  connection.ssl_timeout = @ssl_timeout if @ssl_timeout
  connection.ssl_version = @ssl_version if @ssl_version

  connection.verify_depth = @verify_depth
  connection.verify_mode  = @verify_mode

  if OpenSSL::SSL::VERIFY_PEER == OpenSSL::SSL::VERIFY_NONE and
     not Object.const_defined?(:I_KNOW_THAT_OPENSSL_VERIFY_PEER_EQUALS_VERIFY_NONE_IS_WRONG) then
    warn <<-WARNING
                           !!!SECURITY WARNING!!!

The SSL HTTP connection to:

#{connection.address}:#{connection.port}

                         !!!MAY NOT BE VERIFIED!!!

On your platform your OpenSSL implementation is broken.

There is no difference between the values of VERIFY_NONE and VERIFY_PEER.

This means that attempting to verify the security of SSL connections may not
work.  This exposes you to man-in-the-middle exploits, snooping on the
contents of your connection and other dangers to the security of your data.

To disable this warning define the following constant at top-level in your
application:

I_KNOW_THAT_OPENSSL_VERIFY_PEER_EQUALS_VERIFY_NONE_IS_WRONG = nil

    WARNING
  end

  connection.ca_file = @ca_file if @ca_file
  connection.ca_path = @ca_path if @ca_path

  if @ca_file or @ca_path then
    connection.verify_mode = OpenSSL::SSL::VERIFY_PEER
    connection.verify_callback = @verify_callback if @verify_callback
  end

  if @certificate and @private_key then
    connection.cert = @certificate
    connection.key  = @private_key
  end

  connection.cert_store = if @cert_store then
                            @cert_store
                          else
                            store = OpenSSL::X509::Store.new
                            store.set_default_paths
                            store
                          end
end

#start(http) ⇒ Object

Starts the Net::HTTP connection



725
726
727
728
729
730
731
732
733
734
735
736
737
738
# File 'lib/net/http/persistent.rb', line 725

def start http
  http.set_debug_output @debug_output if @debug_output
  http.open_timeout = @open_timeout if @open_timeout

  http.start

  socket = http.instance_variable_get :@socket

  if socket then # for fakeweb
    @socket_options.each do |option|
      socket.io.setsockopt(*option)
    end
  end
end

#unescape(str) ⇒ Object

URI::unescape wrapper



705
706
707
# File 'lib/net/http/persistent.rb', line 705

def unescape str
  CGI.unescape str if str
end